package web::spider;
use LWP::UserAgent;
use HTTP::Cookies;
use Carp;
use Data::Dumper;
use web::URL;

# Grundgerüst eine webspiders
# folgene Optionen für "new" sind bekannt:
#
# - lwp_opts
#  hashref die an LWP::UserAgent übergeben wird
#
# - user_agent
#  setzen des UserAgent mit dem sich der Spider am Server identifiziert
#
# - connection_timeout
#  Wartezeit bis zum verbiundungsabbruch (default: 30 Sekunden)
#
# - max_file_size
#  maximale dateigröße die heruntergelden wird (default: 250000 byte)
#
# - cookie_file
#  Datei in der die Cookies gespeichert werden (default: ./cookies.txt)
#
# - sleep
#  wartezeit zwischen zwei zugriffen
#
# - sleep_random
#  es wird zusätzlich deine zufällige Zeit zwischen 0 und der zahl gewartet
#
# - url_adding_max
#  mittels "add_again_url" kann man eine schon besuchte url wieder hinzufügen
#  ein wert >0 verhindert das die selbe URL mehr als dieser wert erneut hinzugefügt werden kann
#  das soll endlose rekursioen bei einer fehlerhaften seit verhindern
#  Der defautwert ist 1
#
# - head_code
#  wird ausgeführt nachdem der Heder gelesen wurde.
#  wird 0 (false) zurück geliefert stoppt der Download, bei 1 (true) wird er fortgeführt
#  übergeben wird das web::spider Objeckt und ein HTTP::Headers Objeckt
#
# - body_code
#  wird ausgeführt nachdem der download beendet wurde
#  ist der rückgabewert "false" dann wird das spidern beendet.
#  übergeben wird das web::spider Objeckt und ein HTTP::Response Objeckt
#
# - body_stream_code
#  wird während des downloads ausgeführt
#  nachdem ein Chunk ein "Dateibrocken" geladen wurde wird es ausgeführt
#  der download wird abgebrochen wenn der rückgabe wert "false" ist
#  übergeben wird das web::spider Objeckt und ein HTTP::Response Objeckt und die gerade geladenen Daten.
#
#
# die Idee ist es dass der Spider einem soviel arbeit wie möglich abnimmt,
# aber einem noch die nötige Flexibilität lässt,
# um den Spider so zu gestalten wie man es sich wünscht.

# wichtige metoden:
#
# -> has_next()
#  gibt "true" zurück wenn noch urls geladen werden können
#
# -> check_url()
#  übergeben: ein web::URL objekt oder ein urlstring
#  gibt true zrück wenn die url noch nicht geladen wurde
#
# -> get_deep_now()
#  bei jeder weitern seite die besucht wird wird dieser wert zur vorherigen um eins erhöt
#  damit kann man die augenblickliche rekusrionstiefe feststellen
#
# -> get_seen_list()
#  liste aller besuchten Seiten
#
# -> get_seen_list()
#  liste aler seiten, die noch besucht werden sollen
#
# -> set_head_action($code)
#  siehe "head_code"
#
# -> set_body_action($code)
#  siehe "body_code"
#
# -> set_body_stream_action($code)
#  siehe "body action_code"
#
# -> add_seen_url($url)
#  fügt eine url der "seen" Liste hinzu
#
# -> add_url($url)
#  fügt eine url der "next" Liste hinzu
#
# -> add_again_url($url)
#  fügt eine schonmal besuchte UL der "next" Liste hinzu
#  das funktioniert aber nur so häufig wie es in url_adding_max festgelegt wurde
#
# -> set_user_agent($name)
#  siehe "user_agent"
#
# -> set_proxy ($url, <$protocoll>)
#   einen proxy setzen
#
# -> runn
#  den Spider starten



sub new
{
  my $class=ref($_[0])?ref(shift):shift;

  if(@_ && !@_%2)
  {
    croak "ungerade Anzahl von Elementen übergen, ist kein Hash!";
    return undef;
  }

  my $self={};
  $self->{seen}={};
  $self->{next}=[];

  $self->{now}=undef;
  $self->{base}=undef;

  $self->{lwp}=undef;
  $self->{opts}={@_};

  bless($self,$class);
  $self->init();
  return $self;
}

sub init
{
  my $self=shift;

  my %opts=();
  if($self->_opt('lwp_opts') && ref($self->_opt('lwp_ops')) eq 'HASH')
  { %opts=%{$self->_opt('lwp_ops')}; }
  $self->lwp(LWP::UserAgent->new(%opts));

  $self->lwp->agent($self->_opt('user_agent') || 'Mozilla/5.0');
  $self->lwp->timeout($self->_opt('connection_timeout') || 60);
  #$self->lwp->max_size($self->_opt('max_file_size') || 250_000);
  $self->lwp->protocols_forbidden( [ 'file', 'mailto', 'javascript'] );

  my $cjar = HTTP::Cookies->new(
        file => $self->_opt('cookie_file') || "./cookies.txt",
        autosave => 1,
        max_cookie_size => 4096,
        max_cookies_per_domain => 5,
      );
  $self->lwp->cookie_jar($cjar);

  $self->{opts}->{sleep}=0 unless($self->_opt('sleep'));
  $self->{opts}->{sleep_random}=0 unless($self->_opt('sleep_random'));
  $self->{opts}->{url_adding_max}=1 unless($self->_opt('url_adding_max'));

  $self->{opts}->{deep_max}=10 unless($self->_opt('deep_max'));

  $self->{opts}->{url_ignore}=[] unless($self->_opt('url_ignore'));
  unless(ref($self->_opt('url_ignore')) eq 'ARRAY')
  { $self->{opts}->{url_ignore}=[$self->{opts}->{url_ignore}]; }


  $self->set_head_action() unless($self->_opt('head_code') || ref($self->_opt('head_code')) ne 'CODE');
  $self->set_body_action() unless($self->_opt('body_code') || ref($self->_opt('body_code')) ne 'CODE');
  $self->set_body_stream_action() unless($self->_opt('body_stream_code') || ref($self->_opt('body_stream_code')) ne 'CODE');
}

########################################################################

sub lwp
{
  my $self=shift;
  my $lwp=shift;
  if(!defined($lwp) || ref($lwp) eq 'LWP::UserAgent')
  { return $self->_gs('lwp',$lwp); }
  return undef;
}

sub has_next
{ return scalar(@{shift()->{next}}); }

sub check_url
{
  my $self=shift;
  my $url=shift;
  $url=web::URL->new($url) if($url && !ref($url));
  if($url)
  {
    return 0 unless(ref($url)=~/web::URL/);

    # nicht beliebig tiefe links.
    if($self->_opt('deep_max') > 0)
    { return 0 if($self->get_deep_now() >= $self->_opt('deep_max')); }

    my $forbidden=join('|',@{$self->lwp->protocols_forbidden()});
    return 0 if($url->scheme()=~/^$forbidden$/);

    for my $ignore (@{$self->_opt('url_ignore')})
    { return 0 if($url->abs->as_string()=~/\Q$ignore/); }

    return 1 unless(exists($self->{seen}->{$url->abs()->as_string()}));
  }
  return 0;
}

sub get_base
{
  my $self=shift;
  my $base=$self->_g('base');
  return $base->clone() if($base);
  return undef;
}

sub get_deep_now
{
  my $self=shift;
  my $now=$self->_g('now');
  return $now->deep() if($now);
  return -1;
}

sub get_url_now
{
  my $self=shift;
  my $now=$self->_g('now');
  return $now->clone() if($now);
  return undef;
}

sub get_seen_list
{
  my $self=shift;
  return grep{$self->_g('seen')->{$_}>1}keys(%{$self->_g('seen')});
}

sub get_next_list
{
  my $self=shift;
  #return grep{$_->abs()->as_string()}@{$self->_g('next')};
  return grep{$_->abs($_->parent())}@{$self->_g('next')};
}

########################################################################

sub set_head_action
{
  my $self=shift;
  my $code=shift;
  if($code && ref($code) eq 'CODE')
  {
    $self->{opts}->{head_code}=$code;
    return 1;
  }
  $self->{opts}->{head_code}=undef;
  return 0;
}

sub set_body_action
{
  my $self=shift;
  my $code=shift;
  if($code && ref($code) eq 'CODE')
  {
    $self->{opts}->{body_code}=$code;
    return 1;
  }
  $self->{opts}->{body_code}=undef;
  return 0;
}

sub set_body_stream_action
{
  my $self=shift;
  my $code=shift;
  if($code && ref($code) eq 'CODE')
  {
    $self->{opts}->{body_stream_code}=$code;
    return 1;
  }
  $self->{opts}->{body_stream_code}=undef;
  return 0;
}

sub add_seen_url
{
  my $self=shift;
  my $url=shift;
  if(defined($url))
  {
    if(ref($url))
    {
      $url=$url->abs() if($url->can('abs'));
      $url=$url->as_string() if($url->can('as_string'));
    }

    $self->{seen}->{$url}++;
    return 1;
  }
  return 0;
}

sub add_url
{
  #print Dumper(\@_);
  my $self=shift;
  my $url=shift;
  my $parent=shift;
  my $deep=shift || 0;

  return 0 unless(defined($url));

  my $base=undef;
  $base=$self->_g('now')->abs() if($self->_gs('now'));
  $base=$self->_g('base')->abs() if($self->_gs('base'));
  if($base)
  { $url = $base->new($url); }
  else
  {
    $deep=$self->get_deep_now()+1 unless($deep);
    $url=web::URL->new($url) unless(ref($url) eq 'web::URL');
    $url->deep($deep);
    $url->parent($parent);
  }

  #print Dumper($url);

  if($self->check_url($url))
  {
    push(@{$self->{next}},$url);
    $self->add_seen_url($url);
    return 1;
  }
  return 0;
}

sub add_again_url
{
  my $self=shift;
  my $url=$self->_g('now');
  return 0 unless($url);

  $url=$url->abs()->as_string();
  if(exists($self->{seen}->{$url}) &&
     $self->{seen}->{$url} < $self->_opt('url_adding_max')+2 )
  {
    unshift(@{$self->{next}},$url);
    $self->add_seen_url($url);
    return 1;
  }
  return 0;
}

sub set_user_agent
{
  my $self=shift;
  my $agent=shift;
  if($agent)
  {
    $self->lwp->agent($agent);
    return 1;
  }
  return 0;
}

sub set_proxy
{
  my $self=shift;
  my $proxy=shift;
  my $proto=shift || 'http';

  if($proxy)
  {
    $proxy=web::URL->new($proxy) unless(ref($proxy) eq 'web::URL');
    if($proxy)
    {
      $self->{AGENT}->proxy($proto,$proxy->abs());
      return 1;
    }
  }
  return 0;
}

########################################################################

sub run
{
  my $self=shift;
  my $url=shift;

  # wenn url übergeben eintragen
  $self->add_url($url) if($url);

  #print "START:\n",Dumper($self->{next}),"\n";

  # alle urls abarbeiten
  while($url=shift(@{$self->{next}}))
  {
    #print "HOLE URL ".Dumper($url)."\n";

    $self->_s('base',undef);
    $self->_gs('now',$url);

    my $first=1;
    my $document=undef;
    my $size=0;
    $document='' if($self->_opt('body_code'));

    my $req=HTTP::Request->new('GET',$url->abs());
    $req->header(Referer => $url->parent())if($url->parent());
    my $res = $self->lwp->request( $req, sub
          {
            my ($chunk, $res)=@_;
            eval{ $document.=$chunk if(defined($document)); };
            eval{ $size+=length($chunk); };

            if($first)
            {
              my $code=$self->_opt('head_code');
              if(defined($code) && ref($code) eq 'CODE')
              {
                my $ok=eval{ return $code->($self,$res->headers()) };
                if($@)
                {
                  croak("ERROR runn \"head_code\"\n$@");
                  die($@);
                }
                die('"head_code" returned false') unless($ok);
              }
              $first=0;
            }

            my $code=$self->_opt('body_stream_code');
            if(defined($code) && ref($code) eq 'CODE')
            {
              my $ok=eval{ return $code->($self,$res,$chunk,$size)};
              if($@)
              {
                croak("ERROR runn \"body_stream_code\"\n$@");
                die($@);
              }
              die('"body_stream_code" returned false') unless($ok);
            }


            return 1;
          });
#    my $res = $self->lwp->request($req);
    unless($res)
    {
      carp "$url failed GET!";
      next;
    }

    $self->_s('base',web::URL->new($res->base()));
    $self->_g('base')->deep($url->deep());
    $self->_g('base')->parent($url->parent());

    # wenn body_code gefüllt ist, dann ausführen
    if($self->_opt('body_code'))
    {
      #print "#"x80,"\n",$document,"\n","#"x80,"\n";
      $res->content($document);
      return 1 unless($self->_opt('body_code')->($self,$res));
    }

    $self->add_seen_url($url);
    my $sleep=$self->_opt('sleep') || 0;
    my $sleep_random=int(rand( $self->_opt('sleep_random') || 0 ));
    sleep($sleep+$sleep_random);
  }

  return 1;
}

########################################################################
sub _gs
{
  my $self=shift;
  my $name=shift;
  my $value=shift;
  if(exists($self->{$name}))
  {
    if(defined($value))
    {
      $self->{$name}=$value;
    }
    return $self->{$name};
  }
  return undef;
}

sub _s
{
  my $self=shift;
  my $name=shift;
  my $value=shift;
  if(exists($self->{$name}))
  {
    $self->{$name}=$value;
    return 1;
  }
  return 0;
}

sub _g
{
  my $self=shift;
  my $name=shift;
  return $self->{$name} if(exists($self->{$name}));
  return undef;
}

sub _opt
{
  my $self=shift;
  my $name=shift;
  return $self->_g('opts')->{$name} if(defined($name) && exists($self->_g('opts')->{$name}));
  return undef;
}

########################################################################
########################################################################
1;
